@@ -30,9 +30,11 @@ import signal
 import subprocess
 import traceback
 import errno
+import time
 import re
-
-from Autopkgtest import shellquote_arg, shellquote_cmdl
+import pipes
+import socket
+import shutil
 
 debuglevel = None
 progname = "<VirtSubproc>"
@@ -44,6 +46,8 @@ downtmp = None
 down = None
 downkind = None
 downs = None
+cleaning = False
+in_mainloop = False
 
 
 class Quit:
@@ -53,7 +57,7 @@ class Quit:
         self.m = m
 
 
-class Timeout:
+class Timeout(RuntimeError):
     pass
 
 
@@ -62,6 +66,7 @@ def alarm_handler(*a):
 
 
 def timeout_start(to):
+    signal.signal(signal.SIGALRM, alarm_handler)
     signal.alarm(to)
 
 
@@ -82,7 +87,12 @@ def debug(m):
 
 
 def bomb(m):
-    raise Quit(12, progname + ": failure: %s" % m)
+    if in_mainloop:
+        raise Quit(12, progname + ": failure: %s" % m)
+    else:
+        sys.stderr.write(m)
+        sys.stderr.write('\n')
+        sys.exit(1)
 
 
 def ok():
@@ -180,6 +190,57 @@ def execute(cmd_string, cmd_list=[], downp=False, outp=False, timeout=0):
     return out
 
 
+class timeout:
+    def __init__(self, secs, exit_msg=None):
+        '''Context manager that times out after given number of seconds.
+
+        If exit_msg is given, the program bomb()s with that message,
+        otherwise it raises a Timeout exception.
+        '''
+        self.secs = secs
+        self.exit_msg = exit_msg
+
+    def __enter__(self):
+        timeout_start(self.secs)
+
+    def __exit__(self, type_, value, traceback):
+        timeout_stop()
+        if type_ is Timeout and self.exit_msg:
+            bomb(self.exit_msg)
+            return True
+        return False
+
+
+def get_unix_socket(path):
+    '''Open a connected client socket to given Unix socket with a 5s timeout'''
+
+    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+    with timeout(5, 'Timed out waiting for %s socket\n' % path):
+        while True:
+            try:
+                s.connect(path)
+                break
+            except socket.error:
+                continue
+    return s
+
+
+def expect(sock, search_str, timeout_sec, description=None):
+    debug('expect: "%s"' % search_str)
+    what = '"%s"' % (description or search_str or 'data')
+    out = b''
+    with timeout(timeout_sec,
+                 description and ('timed out waiting for %s' % what) or None):
+        while True:
+            time.sleep(0.1)
+            block = sock.recv(4096)
+            #debug('expect: got block: %s' % block)
+            out += block
+            if search_str is None or search_str in out:
+                debug('expect: found "%s"' % what)
+                break
+
+
 def cmd_open(c, ce):
     global downtmp
     cmdnumargs(c, ce)
@@ -193,7 +254,7 @@ def cmd_open(c, ce):
 
 def downtmp_mktemp():
     d = execute('mktemp -d /tmp/adt-run.XXXXXX', [], downp=True, outp=True)
-    execute('chmod 755', [d], downp=True)
+    execute('chmod 1777', [d], downp=True)
     return d
 
 
@@ -289,7 +350,7 @@ def cmd_execute(c, ce):
 
     for ioe in range(3):
         rune += " %d%s%s" % (ioe, '<>'[ioe > 0],
-                             shellquote_arg(ce[ioe + 2]))
+                             pipes.quote(ce[ioe + 2]))
     if debug_g:
         (tfd, hfd) = m.groups()
         tfd = int(tfd)
@@ -298,13 +359,14 @@ def cmd_execute(c, ce):
 
     rune += '; '
 
-    rune += 'cd %s; ' % shellquote_arg(ce[5])
+    rune += 'cd %s; ' % pipes.quote(ce[5])
 
     for e in envs:
         (en, ev) = map(urllib.unquote, e)
-        rune += "%s=%s " % (en, shellquote_arg(ev))
+        rune += "%s=%s " % (en, pipes.quote(ev))
 
-    rune += 'exec ' + shellquote_cmdl(map(urllib.unquote, ce[1].split(',')))
+    cmdl = map(urllib.unquote, ce[1].split(','))
+    rune += 'exec ' + ' '.join(map(pipes.quote, cmdl))
 
     cmdl = downs['shstring'] + [rune]
 
@@ -329,6 +391,108 @@ def cmd_execute(c, ce):
     return [str(status)]
 
 
+def get_downtmp_host():
+    '''Return host directory of the testbed's downtmp dir, if supported'''
+
+    for cap in caller.hook_capabilities():
+        if cap.startswith('downtmp-host='):
+            return cap.split('=', 1)[1]
+    return None
+
+
+def copytree(src, dst):
+    '''Like shutils.copytree(), but merges with existing dst'''
+
+    if not os.path.exists(dst):
+        shutil.copytree(src, dst, symlinks=True)
+        return
+
+    for f in os.listdir(src):
+        fsrc = os.path.join(src, f)
+        subprocess.check_call(['cp', '-at', dst, fsrc])
+
+
+def copyup_shareddir(tb, host, is_dir, downtmp_host):
+    debug('copyup_shareddir: tb %s, host %s, is_dir %s, downtmp_host %s' % (
+        tb, host, is_dir, downtmp_host))
+
+    timeout_start(copy_timeout)
+    try:
+        tb_tmp = None
+        if tb.startswith(downtmp):
+            # translate into host path
+            tb = downtmp_host + tb[len(downtmp):]
+        else:
+            tb_tmp = os.path.join(downtmp, os.path.basename(host))
+            debug('copyup_shareddir: tb path %s is not already in downtmp, '
+                  'copying to %s' % (tb, tb_tmp))
+            cp = subprocess.Popen(downs['auxverb'] + ['cp', '-a', tb, tb_tmp],
+                                  preexec_fn=preexecfn)
+            cp.communicate()
+            if cp.returncode != 0:
+                bomb('copyup_shareddir: cp -a exited with code %i' %
+                     cp.returncode)
+            # translate into host path
+            tb = os.path.join(downtmp_host, os.path.basename(host))
+
+        if tb == host:
+            tb_tmp = None
+        else:
+            debug('copyup_shareddir: tb(host) %s is not already at '
+                  'destination %s, copying' % (tb, host))
+            if is_dir:
+                copytree(tb, host)
+            else:
+                shutil.copy(tb, host)
+
+        if tb_tmp:
+            (is_dir and shutil.rmtree or os.unlink)(tb_tmp)
+    finally:
+        timeout_stop()
+
+
+def copydown_shareddir(host, tb, is_dir, downtmp_host):
+    debug('copydown_shareddir: host %s, tb %s, is_dir %s, downtmp_host %s' % (
+        host, tb, is_dir, downtmp_host))
+
+    host = os.path.normpath(host)
+    tb = os.path.normpath(tb)
+    downtmp_host = os.path.normpath(downtmp_host)
+
+    timeout_start(copy_timeout)
+    try:
+        host_tmp = None
+        if host.startswith(downtmp_host):
+            # translate into tb path
+            host = downtmp + host[len(downtmp_host):]
+        else:
+            host_tmp = os.path.join(downtmp_host, os.path.basename(tb))
+            if is_dir:
+                shutil.rmtree(host_tmp, ignore_errors=True)
+                shutil.copytree(host, host_tmp, symlinks=True)
+            else:
+                shutil.copy(host, host_tmp)
+            # translate into tb path
+            host = os.path.join(downtmp, os.path.basename(tb))
+
+        if host == tb:
+            host_tmp = None
+        else:
+            subprocess.call(downs['auxverb'] + ['rm', '-rf', tb],
+                            preexec_fn=preexecfn)
+            cp = subprocess.Popen(downs['auxverb'] + ['cp', '-a', host, tb],
+                                  preexec_fn=preexecfn)
+            cp.communicate()
+            if cp.returncode != 0:
+                bomb('copydown_shareddir: cp -a exited with code %i' %
+                     cp.returncode)
+
+        if host_tmp:
+            (is_dir and shutil.rmtree or os.unlink)(host_tmp)
+    finally:
+        timeout_stop()
+
+
 def copyupdown(c, ce, upp):
     cmdnumargs(c, ce, 2)
     if not downtmp:
@@ -345,9 +509,20 @@ def copyupdown(c, ce, upp):
     if dirsp != (sd[1][-1] == '/'):
         bomb("% paths must agree about directoryness"
              " (presence or absence of trailing /)" % wh)
+
+    # if we have a shared directory, we just need to copy it from/to there; in
+    # most cases, it's testbed end is already in the downtmp dir
+    downtmp_host = get_downtmp_host()
+    if downtmp_host:
+        if upp:
+            copyup_shareddir(sd[0], sd[1], dirsp, downtmp_host)
+        else:
+            copydown_shareddir(sd[0], sd[1], dirsp, downtmp_host)
+        return
+
     deststdout = devnull_read
     srcstdin = devnull_read
-    remfileq = shellquote_arg(sd[iremote])
+    remfileq = pipes.quote(sd[iremote])
     if not dirsp:
         rune = 'cat %s%s' % ('><'[upp], remfileq)
         if upp:
@@ -419,9 +594,28 @@ def cmd_copyup(c, ce):
     copyupdown(c, ce, True)
 
 
+def cmd_shell(c, ce):
+    cmdnumargs(c, ce, 4)
+    if not downtmp:
+        bomb("`shell' when not open")
+    try:
+        caller.hook_shell(c[1], c[2], c[3], c[4])
+    except AttributeError:
+        raise FailedCmd(['not supported by virt server'])
+
+
 def command():
     sys.stdout.flush()
-    ce = sys.stdin.readline()
+    while True:
+        try:
+            ce = sys.stdin.readline()
+            break
+        except IOError as e:
+            if e.errno == errno.EAGAIN:
+                time.sleep(0.1)
+                continue
+            else:
+                raise
     if not ce:
         bomb('end of file - caller quit?')
     ce = ce.rstrip().split()
@@ -483,7 +677,7 @@ def error_cleanup():
 
 
 def prepare():
-    global downtmp, cleaning
+    global downtmp
     downtmp = None
 
     def handler(sig, *any):
@@ -493,6 +687,9 @@ def prepare():
 
 
 def mainloop():
+    global in_mainloop
+    in_mainloop = True
+
     try:
         while True:
             command()
@@ -506,10 +703,11 @@ def mainloop():
         print >> sys.stderr, "Unexpected error:"
         traceback.print_exc()
         sys.exit(16)
+    finally:
+        in_mainloop = False
 
 
 def main():
-    signal.signal(signal.SIGALRM, alarm_handler)
     ok()
     prepare()
     mainloop()
